//**************************************************************
//*
//* Pinbelegung
//*    ----------------------------------
//*        1  VDD (+5V)
//*        2  GP5 -out- (shift register clock pin (SH_CP) of 74HC595)
//*        3  GP4 -out- (storage register clock pin (ST_CP) of 74HC595)
//*        4  GP3 -in - (Taster - Messbereich/Binranzeige)
//*        5  GP2 -out- (Data in (DS) of 74HC595)
//*        6  GP1 -out- (testmode pin)
//*        7  AN0 -in - (Stromwandler)
//*        8  VSS (GND)
//*
//**************************************************************
//
// loetmeister.de - 2015, v0.3
//
// Amperemeter + Bargraph (mit Stromwandler/CT - Current Transformer)
//
// PIC 12F675
// Strommessung von 230V Verbrauchern (z.B. PC) ber einen kleinen 'Trafo'
// (Stromwander - current transformer (CT)) und ADC.
// Darstellung der Messwerte mit 8 LEDs als Bargraphanzeige (rot, gelb, grn)
// zur Erkennung von berlast. (16A)
//
// http://www.loetmeister.de/Elektronik/amperemeter_stromwandler+bargraph.htm
//
// Prozessortakt:  4 MHz intern
// Eingangsspannung: 5V

/******************************************************************************/
/* Files to Include                                                           */
/******************************************************************************/

#include <xc.h>         /* XC8 General Include File */

#include <stdint.h>        /* For uint8_t definition */
#include <stdbool.h>       /* For true/false definition */

#include "system.h"        /* System funct/params, like osc/peripheral config */
#include "user.h"          /* User funct/params, such as InitApp */

/******************************************************************************/
/* User Global Variable Declaration                                           */
/******************************************************************************/
signed int AverageI=0, sumI, ADCresult_buf;
unsigned int ADCresult;

signed char offsetI_corretion;
char ADCindex=0;
char timer2=const_timer2;
char Button_timer2=const_Button_timer2;
char waitcounter;

bit ButtonState;
bit programmingDone=0;
bit waiting;
bit CalculateAverage;

// to store the ADC result values
union {unsigned int ADRES;
    unsigned char ad_hi_low[2];
} ADCout;

/******************************************************************************/
/* Main Program                                                               */
/******************************************************************************/
void main(void)
{
    /* Configure the oscillator for the device */
    //ConfigureOscillator(); // no need here, default startup code will do it

    /* Initialize I/O and Peripherals */
    InitApp();

    shiftOut(0b0);  // clear any strange value on the register (as we don't use "Master Reset" pin)

    // start-up light show...
    shiftOut(0b10101010);   // every second LED on
    waitcounter=40;
    waiting=true;
    while (waiting)
        ;

    shiftOut(0b01010101);   // every other second LED on
    waitcounter=80;
    waiting=true;
    while (waiting)
        ;

    shiftOut(0b0);
    show_animation();
    
    
    /* read EEPROM values */
    offsetI_corretion = eeprom_read(0x1);
    if (offsetI_corretion > 20 || offsetI_corretion < -20)
        offsetI_corretion = 0;


    while(1)    // main loop
    {
        sumI = 0;
        di();   // not interrupts in our sampling!
        /* sample sinus waves - 20ms */
        // reset waitcounter and sumI result once, if AC zero crossing happened??
        // save value of waitcounter*2 when it happens again to stop after two half sin waves?
#ifdef TESTMODE
        outTestPin = 1;
#endif
        for (waitcounter = const_waitcounter; waitcounter != 0; waitcounter--)
        {
            // sample input for analog channel 0 (Transformer voltage ~ power consumption)
            ADCON0bits.CHS = 0b00;    // select Analog Channel AN0

            _delay(28); // wait 28 cycles (1/(4MHz/4) = 1s * 28 = 28 s -- 58?
            adc_read();
            ADCout.ad_hi_low[0] = ADRESL;
            ADCout.ad_hi_low[1] = ADRESH;
            
            ADCresult_buf = ADCout.ADRES - const_offsetI;
            
            if (ADCresult_buf < 0)
                ADCresult_buf = ADCresult_buf * -1;
            else
                _delay(6);  // use the same (similar) time in the loop
            
            sumI += ADCresult_buf;
        }
#ifdef TESTMODE
        outTestPin = 0;
#endif
        // calculate the average over the last samples
        sumI = sumI  / const_waitcounter;
        ei();
        
        
        /* average over the last n results. Calculate after evey timer0 roll over*/
        if (CalculateAverage)
        {
            AverageI += (sumI + offsetI_corretion);
            CalculateAverage = false;
            ++ADCindex;
        }
        if (ADCindex > const_ADCsamples)
        {
            ADCindex = 0; // reset when last array element was processed
            ADCresult = AverageI / const_ADCsamples;  // calculate average
            AverageI = 0;
        }


        /* button read section */
        if (inButtonPin != ButtonPressed)
        {
            Button_timer2 = const_Button_timer2; // reset Button_timer2 if not
            programmingDone = false;
        }
        if (Button_timer2 <= (const_Button_timer2 -1) && inButtonPin == ButtonPressed)  // we only want to know when button is pressed
        {
            shiftOut(sumI);
        }
        /* learnButton_timer2 decrement by one at every timer2 role over (4,1Hz) delay is defnied by const_learnButton_timer2  */
        if (Button_timer2 <= 0 && inButtonPin == ButtonPressed && !programmingDone)  // we only want to know when button is pressed
        {            
            // when button was pressed and hold, store the offsetI in the EEPROM
            // PIC12F629/675  devices  have  128 bytes of data EEPROM with an address range from 0h to 7Fh

            // to update the offset, no current should flow through the CT!
            if (sumI <= 20 && sumI >= -20)
            {
                offsetI_corretion = sumI;  // use offset from the last reading. This delta will be used for future calculations
                eeprom_write(0x1, offsetI_corretion);
                // success
                shiftOut(0b00001010); // light up two green LEDs
            }
            else
            {
                // offset not updated
                shiftOut(0b11000000); // light up two red LEDs
            }
            // wait 3 seconds to display the result
            waitcounter=188;
            waiting=true;
            while (waiting)
                ;

            programmingDone = true;  // just saved the value. Don't do it again, until the button was released
        }

    } // while
}

/******************************************************************************/
/* Interrupt routine                                                          */
/* Using T0IF (timer0 interrupt)                                              */
/******************************************************************************/
void interrupt isr(void)
{
    // reset timer0 when interrupt has occured (62,5Hz)
    if (INTCONbits.T0IF)
    {
        TMR0 = const_timer0;    // reset
        --timer2;

        CalculateAverage = true;

        // check if timer2 ticked down (4,1667Hz)
        if (timer2 <= 0)
        {
            --Button_timer2;
            timer2 = const_timer2;  // reset timer2
            
            if (!waiting && !(inButtonPin == ButtonPressed))
                display_bargraph(ADCresult);
        }
        
        if (waiting)
        {
            --waitcounter;
            if (waitcounter == 0)
                waiting = false;
        }

    } //T0IF
    INTCONbits.T0IF = 0;   // clear interrupt flag
}


/******************************************************************************/
/* Shifts out a byte of data, one bit at a time.                              */
/******************************************************************************/
void shiftOut(char byteToSend)
{
    signed char bitPos;
    
    di(); // disable all interrupts

    // turn off the output so the pins don't light up while shifting bits
    outStorageClockPin = 0;
    
    //prepare shift register for bit shifting
    outDataPin = 0;
    outShiftClockPin = 0;
    
    for (bitPos = 7; bitPos >= 0; bitPos--)
    {
        if (byteToSend & (1 << bitPos))   // check current bit and set output accordingly
            outDataPin = 1;
        else
            outDataPin = 0;

        outShiftClockPin = 1;
        outDataPin = 0;
        outShiftClockPin = 0;
    }

    // turn on the output so the LEDs can light up
    outStorageClockPin = 1;

    ei(); // enable all interrupts
}

/******************************************************************************/
/* Takes a char value and lights up the 8 LEDs in binary format               */
/* Data is send to a shift register.                                          */
/******************************************************************************/
void display_binary(char displayValue)
{
    // 30A current -> 1V output, ADC value: 204
//    displayValue = displayValue / 6.8; // no space left for a float valiable...
    shiftOut(displayValue);
}

/******************************************************************************/
void display_bargraph(uint8_t inputValue)
{
    // TODO - add peak hold?
    // 16 Ampere: 0,533V -> 109
    if (inputValue > 103)     shiftOut(0b11111111);
    else if (inputValue > 90) shiftOut(0b01111111);
    else if (inputValue > 77) shiftOut(0b00111111);
    else if (inputValue > 64) shiftOut(0b00011111);
    else if (inputValue > 51) shiftOut(0b00001111);
    else if (inputValue > 38) shiftOut(0b00000111);
    else if (inputValue > 25) shiftOut(0b00000011);
    else if (inputValue > 12) shiftOut(0b00000001);
    else shiftOut(0b00000000);
}

/******************************************************************************/
/* Show a moving LED from low to high                                         */
/******************************************************************************/
void show_animation(void)
{
    char bitPos;

    outStorageClockPin = 0;
    outShiftClockPin = 0;
    outDataPin = 1;
    outShiftClockPin = 1;
    outDataPin = 0;
    
    for (bitPos = 8; bitPos != 0; bitPos--)
    {
        outShiftClockPin = 0;        
        outStorageClockPin = 1;
        outShiftClockPin = 1;
        outStorageClockPin = 0;
        waitcounter=8;
        waiting=true;
        while (waiting)
            ;
    }
}

/******************************************************************************/
// EOF